----- MODULE ASnapshot -----
EXTENDS Naturals, TLC, Sequences
CONSTANT N, K, MAXIT, NONE

(* --algorithm AdaptiveSnapshotAlg
   	
	(* 1. Shared objects: three N x N arrays of atomic registers (we could use
	      3*N*(N+1)/2 registers instead, but the declarations would be more
	      complex). *)

	variables 
		(* Registers used for splitters (2 register for each splitter) *)
		S = [ i \in 1..N, j \in 1..N |-> [ pid |-> {}, flag |-> FALSE ] ],

		(* Registers for storing snapshot values *)
		R = [ i \in 1..N, j \in 1..N |-> [ ts |-> 0, val |-> NONE, snap |-> << >> ] ],

	(* 2. Processes' local variables, most used for returning values from procedures. *)

		spret = [i \in 1..N |-> {}],
		colret = [i \in 1..N |-> {}],
		myreg = [i \in 1..N |-> << >>],
		scanret = [i \in 1..N |-> {}],
		ts = [i \in 1..N |-> 0],

		(* For (partially) testing correctness. *)
		counter = 1;
		define SnapValues == { R[i, j].val : i \in 1..N, j \in 1..N }
		end define;

	(* 3. An implementation of an adaptive atomic snapshot. *)

	procedure splitter(x, y)
	begin
		s1:	S[x, y].pid := self;
		s2:	if S[x, y].flag then spret[self] := "right"; return; end if;
		s3:	S[x, y].flag := TRUE;
		s4:	if S[x, y].pid = self then spret[self] := "stop"; return; end if;
		s5:	spret[self] := "left"; return;
	end procedure;

	procedure collect()
	variables C = << >>, k = 1, emptydiag = FALSE, x, y, flag, reg;
	begin
		c1:	while (~emptydiag) /\ (k <= N) do
				x := 1; emptydiag := TRUE;
		c2:		while x <= k do
					y := k - x + 1;
		c3_Sflag:		flag := S[x, y].flag;
					if flag then 
						emptydiag := FALSE;
		c4_R:				reg := R[x, y];
						if reg.val # NONE then C := Append(C, reg); end if;
					end if;
		c5:			x := x + 1;
				end while;
				k := k + 1;
			end while;
			colret[self] := C;
			return;
	end procedure;
	
	procedure obtain()
	variables x = 1, y = 1;
	begin
		o1:	while TRUE do
		o2:		call splitter(x, y);
		o3:		if spret[self] = "stop" then myreg[self] := << x, y >>; return; end if;
		o4:		if spret[self] = "left" then y := y + 1;
					else x := x + 1;
				end if;
				assert( (x <= K) /\ (y <= K) /\ (x + y <= K + 1) );
			end while;
			return;
	end procedure;

	procedure update(value)
	begin
		u1:	if myreg[self] = << >> then 
		u2:		call obtain();
			end if;
		u3:	ts[self] := ts[self] + 1;
		u4:	call scan();
		u5_R:	R[myreg[self]] := [ ts |-> ts[self], val |-> value, snap |-> scanret[self] ];
			return;
	end procedure;

	procedure scan()
	variables t1, t2, t3, k;
	begin
		s1:	call collect();
		s2:	t1 := colret[self]; t2 := t1;
		s3:	while TRUE do
		s4:		call collect();
		s5:		t3 := colret[self];
				if (Len(t3) = Len(t2)) /\ (t3 = t2) then 
					scanret[self] := [ i \in 1..Len(t3) |-> t3[i].val ]; 
		s6:			return; 
				end if;
		s7:		k := 1;
		s8:		while k <= Len(t3) do
					if ( (k <= Len(t1)) /\ (t3[k].ts >= t1[k].ts + 2) ) \/
					   ( (k > Len(t1)) /\ (t3[k].ts >= 2) ) then
						scanret[self] := t3[k].snap;
						return;
					end if;
		s9:			k := k + 1;
				end while;
		s10:		t2 := t3;
			end while;
	end procedure;

	process Active \in 1..K
	variables s = 1, snap1, snap2, retsnap;
	begin
		tst1:	while s <= MAXIT do
				either
		tst2:			counter := counter + 1;
					call update(counter);
				or
		tst3:			snap1 := SnapValues; (* For testing *)
					call scan();
		tst4:			snap2 := SnapValues; (* For testing *)
					retsnap := scanret[self];
					(* Does not catch all problems, but is better than nothing... *)
					assert( \A i \in 1..Len(retsnap) : (retsnap[i] \in snap1 \cup snap2) /\ (retsnap[i] # NONE) );
				end either;
		tst5:		s := s + 1;
			end while;
	end process
   end algorithm
*)

\* After this line you put the definitions TLA+ operators used by +CAL the algorithm

\* BEGIN TRANSLATION
\* Label s1 of procedure splitter at line 35 col 25 changed to s1_
\* Label s2 of procedure splitter at line 36 col 25 changed to s2_
\* Label s3 of procedure splitter at line 37 col 25 changed to s3_
\* Label s4 of procedure splitter at line 38 col 25 changed to s4_
\* Label s5 of procedure splitter at line 39 col 25 changed to s5_
\* Procedure variable k of procedure collect at line 43 col 30 changed to k_
\* Procedure variable x of procedure collect at line 43 col 56 changed to x_
\* Procedure variable y of procedure collect at line 43 col 59 changed to y_
\* Procedure variable x of procedure obtain at line 64 col 19 changed to x_o
\* Procedure variable y of procedure obtain at line 64 col 26 changed to y_o
VARIABLES S, R, spret, colret, myreg, scanret, ts, counter, pc, stack

(* define statement *)
SnapValues == { R[i, j].val : i \in 1..N, j \in 1..N }

VARIABLES x, y, C, k_, emptydiag, x_, y_, flag, reg, x_o, y_o, value, t1, t2, 
          t3, k, s, snap1, snap2, retsnap

vars == << S, R, spret, colret, myreg, scanret, ts, counter, pc, stack, x, y, 
           C, k_, emptydiag, x_, y_, flag, reg, x_o, y_o, value, t1, t2, t3, 
           k, s, snap1, snap2, retsnap >>

ProcSet == (1..K)

Init == (* Global variables *)
        /\ S = [ i \in 1..N, j \in 1..N |-> [ pid |-> {}, flag |-> FALSE ] ]
        /\ R = [ i \in 1..N, j \in 1..N |-> [ ts |-> 0, val |-> NONE, snap |-> << >> ] ]
        /\ spret = [i \in 1..N |-> {}]
        /\ colret = [i \in 1..N |-> {}]
        /\ myreg = [i \in 1..N |-> << >>]
        /\ scanret = [i \in 1..N |-> {}]
        /\ ts = [i \in 1..N |-> 0]
        /\ counter = 1
        (* Procedure splitter *)
        /\ x = [ self \in ProcSet |-> {}]
        /\ y = [ self \in ProcSet |-> {}]
        (* Procedure collect *)
        /\ C = [ self \in ProcSet |-> << >>]
        /\ k_ = [ self \in ProcSet |-> 1]
        /\ emptydiag = [ self \in ProcSet |-> FALSE]
        /\ x_ = [ self \in ProcSet |-> {}]
        /\ y_ = [ self \in ProcSet |-> {}]
        /\ flag = [ self \in ProcSet |-> {}]
        /\ reg = [ self \in ProcSet |-> {}]
        (* Procedure obtain *)
        /\ x_o = [ self \in ProcSet |-> 1]
        /\ y_o = [ self \in ProcSet |-> 1]
        (* Procedure update *)
        /\ value = [ self \in ProcSet |-> {}]
        (* Procedure scan *)
        /\ t1 = [ self \in ProcSet |-> {}]
        /\ t2 = [ self \in ProcSet |-> {}]
        /\ t3 = [ self \in ProcSet |-> {}]
        /\ k = [ self \in ProcSet |-> {}]
        (* Process Active *)
        /\ s = [self \in 1..K |-> 1]
        /\ snap1 = [self \in 1..K |-> {}]
        /\ snap2 = [self \in 1..K |-> {}]
        /\ retsnap = [self \in 1..K |-> {}]
        /\ stack = [self \in ProcSet |-> << >>]
        /\ pc = [self \in ProcSet |-> CASE self \in 1..K -> "tst1"]

s1_(self) == /\ pc[self] = "s1_"
             /\ S' = [S EXCEPT ![x[self], y[self]].pid = self]
             /\ pc' = [pc EXCEPT ![self] = "s2_"]
             /\ UNCHANGED << R, spret, colret, myreg, scanret, ts, counter, 
                             stack, x, y, C, k_, emptydiag, x_, y_, flag, reg, 
                             x_o, y_o, value, t1, t2, t3, k, s, snap1, snap2, 
                             retsnap >>

s2_(self) == /\ pc[self] = "s2_"
             /\ IF S[x[self], y[self]].flag
                   THEN /\ spret' = [spret EXCEPT ![self] = "right"]
                        /\ pc' = [pc EXCEPT ![self] = Head(stack[self]).pc]
                        /\ x' = [x EXCEPT ![self] = Head(stack[self]).x]
                        /\ y' = [y EXCEPT ![self] = Head(stack[self]).y]
                        /\ stack' = [stack EXCEPT 
                                       ![self] = Tail(stack[self])]
                   ELSE /\ pc' = [pc EXCEPT ![self] = "s3_"]
                        /\ UNCHANGED << spret, stack, x, y >>
             /\ UNCHANGED << S, R, colret, myreg, scanret, ts, counter, C, k_, 
                             emptydiag, x_, y_, flag, reg, x_o, y_o, value, t1, 
                             t2, t3, k, s, snap1, snap2, retsnap >>

s3_(self) == /\ pc[self] = "s3_"
             /\ S' = [S EXCEPT ![x[self], y[self]].flag = TRUE]
             /\ pc' = [pc EXCEPT ![self] = "s4_"]
             /\ UNCHANGED << R, spret, colret, myreg, scanret, ts, counter, 
                             stack, x, y, C, k_, emptydiag, x_, y_, flag, reg, 
                             x_o, y_o, value, t1, t2, t3, k, s, snap1, snap2, 
                             retsnap >>

s4_(self) == /\ pc[self] = "s4_"
             /\ IF S[x[self], y[self]].pid = self
                   THEN /\ spret' = [spret EXCEPT ![self] = "stop"]
                        /\ pc' = [pc EXCEPT ![self] = Head(stack[self]).pc]
                        /\ x' = [x EXCEPT ![self] = Head(stack[self]).x]
                        /\ y' = [y EXCEPT ![self] = Head(stack[self]).y]
                        /\ stack' = [stack EXCEPT 
                                       ![self] = Tail(stack[self])]
                   ELSE /\ pc' = [pc EXCEPT ![self] = "s5_"]
                        /\ UNCHANGED << spret, stack, x, y >>
             /\ UNCHANGED << S, R, colret, myreg, scanret, ts, counter, C, k_, 
                             emptydiag, x_, y_, flag, reg, x_o, y_o, value, t1, 
                             t2, t3, k, s, snap1, snap2, retsnap >>

s5_(self) == /\ pc[self] = "s5_"
             /\ spret' = [spret EXCEPT ![self] = "left"]
             /\ pc' = [pc EXCEPT ![self] = Head(stack[self]).pc]
             /\ x' = [x EXCEPT ![self] = Head(stack[self]).x]
             /\ y' = [y EXCEPT ![self] = Head(stack[self]).y]
             /\ stack' = [stack EXCEPT ![self] = Tail(stack[self])]
             /\ UNCHANGED << S, R, colret, myreg, scanret, ts, counter, C, k_, 
                             emptydiag, x_, y_, flag, reg, x_o, y_o, value, t1, 
                             t2, t3, k, s, snap1, snap2, retsnap >>

splitter(self) == s1_(self) \/ s2_(self) \/ s3_(self) \/ s4_(self)
                     \/ s5_(self)

c1(self) == /\ pc[self] = "c1"
            /\ IF (~emptydiag[self]) /\ (k_[self] <= N)
                  THEN /\ x_' = [x_ EXCEPT ![self] = 1]
                       /\ emptydiag' = [emptydiag EXCEPT 
                                          ![self] = TRUE]
                       /\ pc' = [pc EXCEPT ![self] = "c2"]
                       /\ UNCHANGED << colret, stack, C, k_, y_, flag, reg >>
                  ELSE /\ colret' = [colret EXCEPT ![self] = C[self]]
                       /\ pc' = [pc EXCEPT ![self] = Head(stack[self]).pc]
                       /\ C' = [C EXCEPT ![self] = Head(stack[self]).C]
                       /\ k_' = [k_ EXCEPT ![self] = Head(stack[self]).k_]
                       /\ emptydiag' = [emptydiag EXCEPT 
                                          ![self] = Head(stack[self]).emptydiag]
                       /\ x_' = [x_ EXCEPT ![self] = Head(stack[self]).x_]
                       /\ y_' = [y_ EXCEPT ![self] = Head(stack[self]).y_]
                       /\ flag' = [flag EXCEPT 
                                     ![self] = Head(stack[self]).flag]
                       /\ reg' = [reg EXCEPT ![self] = Head(stack[self]).reg]
                       /\ stack' = [stack EXCEPT 
                                      ![self] = Tail(stack[self])]
            /\ UNCHANGED << S, R, spret, myreg, scanret, ts, counter, x, y, 
                            x_o, y_o, value, t1, t2, t3, k, s, snap1, snap2, 
                            retsnap >>

c2(self) == /\ pc[self] = "c2"
            /\ IF x_[self] <= k_[self]
                  THEN /\ y_' = [y_ EXCEPT ![self] = k_[self] - x_[self] + 1]
                       /\ pc' = [pc EXCEPT ![self] = "c3_Sflag"]
                       /\ UNCHANGED k_
                  ELSE /\ k_' = [k_ EXCEPT ![self] = k_[self] + 1]
                       /\ pc' = [pc EXCEPT ![self] = "c1"]
                       /\ UNCHANGED y_
            /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                            stack, x, y, C, emptydiag, x_, flag, reg, x_o, y_o, 
                            value, t1, t2, t3, k, s, snap1, snap2, retsnap >>

c3_Sflag(self) == /\ pc[self] = "c3_Sflag"
                  /\ flag' = [flag EXCEPT ![self] = S[x_[self], y_[self]].flag]
                  /\ IF flag'[self]
                        THEN /\ emptydiag' = [emptydiag EXCEPT 
                                                ![self] = FALSE]
                             /\ pc' = [pc EXCEPT 
                                         ![self] = "c4_R"]
                        ELSE /\ pc' = [pc EXCEPT 
                                         ![self] = "c5"]
                             /\ UNCHANGED emptydiag
                  /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, 
                                  counter, stack, x, y, C, k_, x_, y_, reg, 
                                  x_o, y_o, value, t1, t2, t3, k, s, snap1, 
                                  snap2, retsnap >>

c4_R(self) == /\ pc[self] = "c4_R"
              /\ reg' = [reg EXCEPT ![self] = R[x_[self], y_[self]]]
              /\ IF reg'[self].val # NONE
                    THEN /\ C' = [C EXCEPT ![self] = Append(C[self], reg'[self])]
                    ELSE /\ TRUE
                         /\ UNCHANGED C
              /\ pc' = [pc EXCEPT ![self] = "c5"]
              /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                              stack, x, y, k_, emptydiag, x_, y_, flag, x_o, 
                              y_o, value, t1, t2, t3, k, s, snap1, snap2, 
                              retsnap >>

c5(self) == /\ pc[self] = "c5"
            /\ x_' = [x_ EXCEPT ![self] = x_[self] + 1]
            /\ pc' = [pc EXCEPT ![self] = "c2"]
            /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                            stack, x, y, C, k_, emptydiag, y_, flag, reg, x_o, 
                            y_o, value, t1, t2, t3, k, s, snap1, snap2, 
                            retsnap >>

collect(self) == c1(self) \/ c2(self) \/ c3_Sflag(self) \/ c4_R(self)
                    \/ c5(self)

o1(self) == /\ pc[self] = "o1"
            /\ pc' = [pc EXCEPT ![self] = "o2"]
            /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                            stack, x, y, C, k_, emptydiag, x_, y_, flag, reg, 
                            x_o, y_o, value, t1, t2, t3, k, s, snap1, snap2, 
                            retsnap >>

o2(self) == /\ pc[self] = "o2"
            /\ /\ stack' = [stack EXCEPT ![self] = << [ procedure |->  "splitter",
                                                        pc        |->  "o3",
                                                        x         |->  x[self],
                                                        y         |->  y[self] ] >>
                                                    \o stack[self]]
               /\ x' = [x EXCEPT ![self] = x_o[self]]
               /\ y' = [y EXCEPT ![self] = y_o[self]]
            /\ pc' = [pc EXCEPT ![self] = "s1_"]
            /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                            C, k_, emptydiag, x_, y_, flag, reg, x_o, y_o, 
                            value, t1, t2, t3, k, s, snap1, snap2, retsnap >>

o3(self) == /\ pc[self] = "o3"
            /\ IF spret[self] = "stop"
                  THEN /\ myreg' = [myreg EXCEPT ![self] = << x_o[self], y_o[self] >>]
                       /\ pc' = [pc EXCEPT ![self] = Head(stack[self]).pc]
                       /\ x_o' = [x_o EXCEPT ![self] = Head(stack[self]).x_o]
                       /\ y_o' = [y_o EXCEPT ![self] = Head(stack[self]).y_o]
                       /\ stack' = [stack EXCEPT 
                                      ![self] = Tail(stack[self])]
                  ELSE /\ pc' = [pc EXCEPT ![self] = "o4"]
                       /\ UNCHANGED << myreg, stack, x_o, y_o >>
            /\ UNCHANGED << S, R, spret, colret, scanret, ts, counter, x, y, C, 
                            k_, emptydiag, x_, y_, flag, reg, value, t1, t2, 
                            t3, k, s, snap1, snap2, retsnap >>

o4(self) == /\ pc[self] = "o4"
            /\ IF spret[self] = "left"
                  THEN /\ y_o' = [y_o EXCEPT ![self] = y_o[self] + 1]
                       /\ UNCHANGED x_o
                  ELSE /\ x_o' = [x_o EXCEPT ![self] = x_o[self] + 1]
                       /\ UNCHANGED y_o
            /\ Assert(( (x_o'[self] <= K) /\ (y_o'[self] <= K) /\ (x_o'[self] + y_o'[self] <= K + 1) ), 
                      "Failure of assertion at line line 72, column 33.")
            /\ pc' = [pc EXCEPT ![self] = "o1"]
            /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                            stack, x, y, C, k_, emptydiag, x_, y_, flag, reg, 
                            value, t1, t2, t3, k, s, snap1, snap2, retsnap >>

obtain(self) == o1(self) \/ o2(self) \/ o3(self) \/ o4(self)

u1(self) == /\ pc[self] = "u1"
            /\ IF myreg[self] = << >>
                  THEN /\ pc' = [pc EXCEPT ![self] = "u2"]
                  ELSE /\ pc' = [pc EXCEPT ![self] = "u3"]
            /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                            stack, x, y, C, k_, emptydiag, x_, y_, flag, reg, 
                            x_o, y_o, value, t1, t2, t3, k, s, snap1, snap2, 
                            retsnap >>

u2(self) == /\ pc[self] = "u2"
            /\ stack' = [stack EXCEPT ![self] = << [ procedure |->  "obtain",
                                                     pc        |->  "u3",
                                                     x_o       |->  x_o[self],
                                                     y_o       |->  y_o[self] ] >>
                                                 \o stack[self]]
            /\ x_o' = [x_o EXCEPT ![self] = 1]
            /\ y_o' = [y_o EXCEPT ![self] = 1]
            /\ pc' = [pc EXCEPT ![self] = "o1"]
            /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                            x, y, C, k_, emptydiag, x_, y_, flag, reg, value, 
                            t1, t2, t3, k, s, snap1, snap2, retsnap >>

u3(self) == /\ pc[self] = "u3"
            /\ ts' = [ts EXCEPT ![self] = ts[self] + 1]
            /\ pc' = [pc EXCEPT ![self] = "u4"]
            /\ UNCHANGED << S, R, spret, colret, myreg, scanret, counter, 
                            stack, x, y, C, k_, emptydiag, x_, y_, flag, reg, 
                            x_o, y_o, value, t1, t2, t3, k, s, snap1, snap2, 
                            retsnap >>

u4(self) == /\ pc[self] = "u4"
            /\ stack' = [stack EXCEPT ![self] = << [ procedure |->  "scan",
                                                     pc        |->  "u5_R",
                                                     t1        |->  t1[self],
                                                     t2        |->  t2[self],
                                                     t3        |->  t3[self],
                                                     k         |->  k[self] ] >>
                                                 \o stack[self]]
            /\ t1' = [t1 EXCEPT ![self] = {}]
            /\ t2' = [t2 EXCEPT ![self] = {}]
            /\ t3' = [t3 EXCEPT ![self] = {}]
            /\ k' = [k EXCEPT ![self] = {}]
            /\ pc' = [pc EXCEPT ![self] = "s1"]
            /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                            x, y, C, k_, emptydiag, x_, y_, flag, reg, x_o, 
                            y_o, value, s, snap1, snap2, retsnap >>

u5_R(self) == /\ pc[self] = "u5_R"
              /\ R' = [R EXCEPT ![myreg[self]] = [ ts |-> ts[self], val |-> value[self], snap |-> scanret[self] ]]
              /\ pc' = [pc EXCEPT ![self] = Head(stack[self]).pc]
              /\ value' = [value EXCEPT ![self] = Head(stack[self]).value]
              /\ stack' = [stack EXCEPT ![self] = Tail(stack[self])]
              /\ UNCHANGED << S, spret, colret, myreg, scanret, ts, counter, x, 
                              y, C, k_, emptydiag, x_, y_, flag, reg, x_o, y_o, 
                              t1, t2, t3, k, s, snap1, snap2, retsnap >>

update(self) == u1(self) \/ u2(self) \/ u3(self) \/ u4(self) \/ u5_R(self)

s1(self) == /\ pc[self] = "s1"
            /\ stack' = [stack EXCEPT ![self] = << [ procedure |->  "collect",
                                                     pc        |->  "s2",
                                                     C         |->  C[self],
                                                     k_        |->  k_[self],
                                                     emptydiag |->  emptydiag[self],
                                                     x_        |->  x_[self],
                                                     y_        |->  y_[self],
                                                     flag      |->  flag[self],
                                                     reg       |->  reg[self] ] >>
                                                 \o stack[self]]
            /\ C' = [C EXCEPT ![self] = << >>]
            /\ k_' = [k_ EXCEPT ![self] = 1]
            /\ emptydiag' = [emptydiag EXCEPT 
                               ![self] = FALSE]
            /\ x_' = [x_ EXCEPT ![self] = {}]
            /\ y_' = [y_ EXCEPT ![self] = {}]
            /\ flag' = [flag EXCEPT ![self] = {}]
            /\ reg' = [reg EXCEPT ![self] = {}]
            /\ pc' = [pc EXCEPT ![self] = "c1"]
            /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                            x, y, x_o, y_o, value, t1, t2, t3, k, s, snap1, 
                            snap2, retsnap >>

s2(self) == /\ pc[self] = "s2"
            /\ t1' = [t1 EXCEPT ![self] = colret[self]]
            /\ t2' = [t2 EXCEPT ![self] = t1'[self]]
            /\ pc' = [pc EXCEPT ![self] = "s3"]
            /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                            stack, x, y, C, k_, emptydiag, x_, y_, flag, reg, 
                            x_o, y_o, value, t3, k, s, snap1, snap2, retsnap >>

s3(self) == /\ pc[self] = "s3"
            /\ pc' = [pc EXCEPT ![self] = "s4"]
            /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                            stack, x, y, C, k_, emptydiag, x_, y_, flag, reg, 
                            x_o, y_o, value, t1, t2, t3, k, s, snap1, snap2, 
                            retsnap >>

s4(self) == /\ pc[self] = "s4"
            /\ stack' = [stack EXCEPT ![self] = << [ procedure |->  "collect",
                                                     pc        |->  "s5",
                                                     C         |->  C[self],
                                                     k_        |->  k_[self],
                                                     emptydiag |->  emptydiag[self],
                                                     x_        |->  x_[self],
                                                     y_        |->  y_[self],
                                                     flag      |->  flag[self],
                                                     reg       |->  reg[self] ] >>
                                                 \o stack[self]]
            /\ C' = [C EXCEPT ![self] = << >>]
            /\ k_' = [k_ EXCEPT ![self] = 1]
            /\ emptydiag' = [emptydiag EXCEPT 
                               ![self] = FALSE]
            /\ x_' = [x_ EXCEPT ![self] = {}]
            /\ y_' = [y_ EXCEPT ![self] = {}]
            /\ flag' = [flag EXCEPT ![self] = {}]
            /\ reg' = [reg EXCEPT ![self] = {}]
            /\ pc' = [pc EXCEPT ![self] = "c1"]
            /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                            x, y, x_o, y_o, value, t1, t2, t3, k, s, snap1, 
                            snap2, retsnap >>

s5(self) == /\ pc[self] = "s5"
            /\ t3' = [t3 EXCEPT ![self] = colret[self]]
            /\ IF (Len(t3'[self]) = Len(t2[self])) /\ (t3'[self] = t2[self])
                  THEN /\ scanret' = [scanret EXCEPT ![self] = [ i \in 1..Len(t3'[self]) |-> t3'[self][i].val ]]
                       /\ pc' = [pc EXCEPT ![self] = "s6"]
                  ELSE /\ pc' = [pc EXCEPT ![self] = "s7"]
                       /\ UNCHANGED scanret
            /\ UNCHANGED << S, R, spret, colret, myreg, ts, counter, stack, x, 
                            y, C, k_, emptydiag, x_, y_, flag, reg, x_o, y_o, 
                            value, t1, t2, k, s, snap1, snap2, retsnap >>

s6(self) == /\ pc[self] = "s6"
            /\ pc' = [pc EXCEPT ![self] = Head(stack[self]).pc]
            /\ t1' = [t1 EXCEPT ![self] = Head(stack[self]).t1]
            /\ t2' = [t2 EXCEPT ![self] = Head(stack[self]).t2]
            /\ t3' = [t3 EXCEPT ![self] = Head(stack[self]).t3]
            /\ k' = [k EXCEPT ![self] = Head(stack[self]).k]
            /\ stack' = [stack EXCEPT ![self] = Tail(stack[self])]
            /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                            x, y, C, k_, emptydiag, x_, y_, flag, reg, x_o, 
                            y_o, value, s, snap1, snap2, retsnap >>

s7(self) == /\ pc[self] = "s7"
            /\ k' = [k EXCEPT ![self] = 1]
            /\ pc' = [pc EXCEPT ![self] = "s8"]
            /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                            stack, x, y, C, k_, emptydiag, x_, y_, flag, reg, 
                            x_o, y_o, value, t1, t2, t3, s, snap1, snap2, 
                            retsnap >>

s8(self) == /\ pc[self] = "s8"
            /\ IF k[self] <= Len(t3[self])
                  THEN /\ IF ( (k[self] <= Len(t1[self])) /\ (t3[self][k[self]].ts >= t1[self][k[self]].ts + 2) ) \/
                             ( (k[self] > Len(t1[self])) /\ (t3[self][k[self]].ts >= 2) )
                             THEN /\ scanret' = [scanret EXCEPT ![self] = t3[self][k[self]].snap]
                                  /\ pc' = [pc EXCEPT 
                                              ![self] = Head(stack[self]).pc]
                                  /\ t1' = [t1 EXCEPT 
                                              ![self] = Head(stack[self]).t1]
                                  /\ t2' = [t2 EXCEPT 
                                              ![self] = Head(stack[self]).t2]
                                  /\ t3' = [t3 EXCEPT 
                                              ![self] = Head(stack[self]).t3]
                                  /\ k' = [k EXCEPT 
                                             ![self] = Head(stack[self]).k]
                                  /\ stack' = [stack EXCEPT 
                                                 ![self] = Tail(stack[self])]
                             ELSE /\ pc' = [pc EXCEPT 
                                              ![self] = "s9"]
                                  /\ UNCHANGED << scanret, stack, t1, t2, t3, 
                                                  k >>
                  ELSE /\ pc' = [pc EXCEPT ![self] = "s10"]
                       /\ UNCHANGED << scanret, stack, t1, t2, t3, k >>
            /\ UNCHANGED << S, R, spret, colret, myreg, ts, counter, x, y, C, 
                            k_, emptydiag, x_, y_, flag, reg, x_o, y_o, value, 
                            s, snap1, snap2, retsnap >>

s9(self) == /\ pc[self] = "s9"
            /\ k' = [k EXCEPT ![self] = k[self] + 1]
            /\ pc' = [pc EXCEPT ![self] = "s8"]
            /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                            stack, x, y, C, k_, emptydiag, x_, y_, flag, reg, 
                            x_o, y_o, value, t1, t2, t3, s, snap1, snap2, 
                            retsnap >>

s10(self) == /\ pc[self] = "s10"
             /\ t2' = [t2 EXCEPT ![self] = t3[self]]
             /\ pc' = [pc EXCEPT ![self] = "s3"]
             /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                             stack, x, y, C, k_, emptydiag, x_, y_, flag, reg, 
                             x_o, y_o, value, t1, t3, k, s, snap1, snap2, 
                             retsnap >>

scan(self) == s1(self) \/ s2(self) \/ s3(self) \/ s4(self) \/ s5(self)
                 \/ s6(self) \/ s7(self) \/ s8(self) \/ s9(self) \/ s10(self)

tst1(self) == /\ pc[self] = "tst1"
              /\ IF s[self] <= MAXIT
                    THEN /\ \/ /\ pc' = [pc EXCEPT 
                                           ![self] = "tst2"]
                            \/ /\ pc' = [pc EXCEPT 
                                           ![self] = "tst3"]
                    ELSE /\ pc' = [pc EXCEPT ![self] = "Done"]
              /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                              stack, x, y, C, k_, emptydiag, x_, y_, flag, reg, 
                              x_o, y_o, value, t1, t2, t3, k, s, snap1, snap2, 
                              retsnap >>

tst5(self) == /\ pc[self] = "tst5"
              /\ s' = [s EXCEPT ![self] = s[self] + 1]
              /\ pc' = [pc EXCEPT ![self] = "tst1"]
              /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                              stack, x, y, C, k_, emptydiag, x_, y_, flag, reg, 
                              x_o, y_o, value, t1, t2, t3, k, snap1, snap2, 
                              retsnap >>

tst2(self) == /\ pc[self] = "tst2"
              /\ counter' = counter + 1
              /\ /\ stack' = [stack EXCEPT ![self] = << [ procedure |->  "update",
                                                          pc        |->  "tst5",
                                                          value     |->  value[self] ] >>
                                                      \o stack[self]]
                 /\ value' = [value EXCEPT ![self] = counter']
              /\ pc' = [pc EXCEPT ![self] = "u1"]
              /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, x, y, C, 
                              k_, emptydiag, x_, y_, flag, reg, x_o, y_o, t1, 
                              t2, t3, k, s, snap1, snap2, retsnap >>

tst3(self) == /\ pc[self] = "tst3"
              /\ snap1' = [snap1 EXCEPT ![self] = SnapValues]
              /\ stack' = [stack EXCEPT ![self] = << [ procedure |->  "scan",
                                                       pc        |->  "tst4",
                                                       t1        |->  t1[self],
                                                       t2        |->  t2[self],
                                                       t3        |->  t3[self],
                                                       k         |->  k[self] ] >>
                                                   \o stack[self]]
              /\ t1' = [t1 EXCEPT ![self] = {}]
              /\ t2' = [t2 EXCEPT ![self] = {}]
              /\ t3' = [t3 EXCEPT ![self] = {}]
              /\ k' = [k EXCEPT ![self] = {}]
              /\ pc' = [pc EXCEPT ![self] = "s1"]
              /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                              x, y, C, k_, emptydiag, x_, y_, flag, reg, x_o, 
                              y_o, value, s, snap2, retsnap >>

tst4(self) == /\ pc[self] = "tst4"
              /\ snap2' = [snap2 EXCEPT ![self] = SnapValues]
              /\ retsnap' = [retsnap EXCEPT ![self] = scanret[self]]
              /\ Assert(( \A i \in 1..Len(retsnap'[self]) : (retsnap'[self][i] \in snap1[self] \cup snap2'[self]) /\ (retsnap'[self][i] # NONE) ), 
                        "Failure of assertion at line line 126, column 41.")
              /\ pc' = [pc EXCEPT ![self] = "tst5"]
              /\ UNCHANGED << S, R, spret, colret, myreg, scanret, ts, counter, 
                              stack, x, y, C, k_, emptydiag, x_, y_, flag, reg, 
                              x_o, y_o, value, t1, t2, t3, k, s, snap1 >>

Active(self) == tst1(self) \/ tst5(self) \/ tst2(self) \/ tst3(self)
                   \/ tst4(self)

Next == (\E self \in ProcSet:  \/ splitter(self) \/ collect(self)
                               \/ obtain(self) \/ update(self) \/ scan(self))
           \/ (\E self \in 1..K: Active(self))
           \/ (* Disjunct to prevent deadlock on termination *)
              (\A self \in ProcSet: pc[self] = "Done" /\ UNCHANGED vars)

Spec == Init /\ [][Next]_vars

Termination == <>(\A self \in ProcSet: pc[self] = "Done")

\* END TRANSLATION

\* After this line you put the invariants you want to check against (in TLA+)

=====
